import contextlib
import os
import sys
import tempfile


@contextlib.contextmanager
def suppress():
    original_stdout_fd = os.dup(sys.stdout.fileno())
    original_stderr_fd = os.dup(sys.stderr.fileno())

    with tempfile.TemporaryFile(mode="w+b") as temp_stdout, tempfile.TemporaryFile(
        mode="w+b"
    ) as temp_stderr:
        os.dup2(temp_stdout.fileno(), sys.stdout.fileno())
        os.dup2(temp_stderr.fileno(), sys.stderr.fileno())

        try:
            yield
        finally:
            sys.stdout.flush()
            sys.stderr.flush()

            os.dup2(original_stdout_fd, sys.stdout.fileno())
            os.dup2(original_stderr_fd, sys.stderr.fileno())

            os.close(original_stdout_fd)
            os.close(original_stderr_fd)



import dataclasses
import os

import gdown
from loguru import logger


@dataclasses.dataclass
class Blob:
    url: str
    hash: str

    @property
    def path(self):
        return os.path.expanduser(f"~/.cache/osam/models/blobs/{self.hash}")

    @property
    def size(self):
        if os.path.exists(self.path):
            return os.stat(self.path).st_size
        else:
            return None

    @property
    def modified_at(self):
        if os.path.exists(self.path):
            return os.stat(self.path).st_mtime
        else:
            return None

    def pull(self):
        gdown.cached_download(url=self.url, path=self.path, hash=self.hash)

    def remove(self):
        if os.path.exists(self.path):
            logger.debug("Removing blob {path!r}", path=self.path)
            os.remove(self.path)
        else:
            logger.warning("Blob {path!r} not found", path=self.path)



import warnings

from ._annotation import Annotation  # noqa: F401
from ._blob import Blob  # noqa: F401
from ._bounding_box import BoundingBox  # noqa: F401
from ._generate import GenerateRequest  # noqa: F401
from ._generate import GenerateResponse  # noqa: F401
from ._image_embedding import ImageEmbedding  # noqa: F401
from ._model import Model  # noqa: F401
from ._prompt import Prompt  # noqa: F401


class ModelBase(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warning(
            "This class is deprecated, use `Model` instead.", DeprecationWarning
        )


class ModelBlob(Blob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warning(
            "This class is deprecated, use `Blob` instead.", DeprecationWarning
        )



from typing import List

import numpy as np
import pydantic


class ImageEmbedding(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    original_height: int
    original_width: int
    embedding: np.ndarray

    @pydantic.field_validator("embedding")
    def validate_embedding(cls, embedding):
        if embedding.ndim != 3:
            raise ValueError(
                "embedding must be 3-dimensional: (embedding_dim, height, width)"
            )
        return embedding

    @pydantic.field_serializer("embedding")
    def serialize_embedding(self, embedding: np.ndarray) -> List[List[List[float]]]:
        return embedding.tolist()



import base64
import io

import numpy as np
import PIL.Image


def image_ndarray_to_b64data(ndarray):
    pil = PIL.Image.fromarray(ndarray)
    f = io.BytesIO()
    pil.save(f, format="PNG")
    data = f.getvalue()
    return base64.b64encode(data).decode("utf-8")


def image_b64data_to_ndarray(b64data):
    data = base64.b64decode(b64data)
    pil = PIL.Image.open(io.BytesIO(data))
    return np.asarray(pil)



from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pydantic

from .. import _json
from ._annotation import Annotation
from ._image_embedding import ImageEmbedding
from ._prompt import Prompt


class GenerateRequest(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    model: str
    image_embedding: Optional[ImageEmbedding] = pydantic.Field(default=None)
    image: Optional[np.ndarray] = pydantic.Field(default=None)
    prompt: Optional[Prompt] = pydantic.Field(default=None)
    annotations: Optional[List[Annotation]] = pydantic.Field(default=None)

    @pydantic.field_validator("image", mode="before")
    @classmethod
    def validate_image(
        cls, image: Optional[Union[str, np.ndarray]]
    ) -> Optional[np.ndarray]:
        if isinstance(image, str):
            return _json.image_b64data_to_ndarray(b64data=image)
        return image


class GenerateResponse(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    model: str
    image_embedding: Optional[ImageEmbedding] = pydantic.Field(default=None)
    annotations: List[Annotation]



from typing import Optional

import numpy as np
import pydantic

from .. import _json
from ._bounding_box import BoundingBox


class Annotation(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    bounding_box: Optional[BoundingBox] = pydantic.Field(default=None)
    text: Optional[str] = pydantic.Field(default=None)
    score: Optional[float] = pydantic.Field(default=None)
    mask: Optional[np.ndarray] = pydantic.Field(default=None)

    @pydantic.field_validator("mask")
    def validate_mask(cls, mask: Optional[np.ndarray]) -> Optional[np.ndarray]:
        if mask is None:
            return None
        if mask.dtype != bool:
            raise ValueError("mask must be boolean arrays")
        return mask

    @pydantic.field_serializer("mask")
    def serialize_mask(self, mask: Optional[np.ndarray]) -> Optional[str]:
        return _json.image_ndarray_to_b64data(ndarray=mask.view(np.uint8) * 255)



from typing import List
from typing import Optional
from typing import Type

from loguru import logger

from . import types

running_model: Optional[types.Model] = None
registered_model_types: List[Type] = []


def register_model_type(model: Type) -> None:
    global registered_model_types

    for registered_model in registered_model_types:
        if registered_model.name == model.name:
            logger.warning(f"Model {model.name!r} is already registered")
            return
    else:
        registered_model_types.append(model)


def get_model_type_by_name(name: str) -> Type:
    model_name: str
    if ":" in name:
        model_name = name
    else:
        model_name = f"{name}:latest"

    for cls in registered_model_types:
        if cls.name == model_name:
            break
    else:
        raise ValueError(f"Model {name!r} not found.")
    return cls


def generate(request: types.GenerateRequest) -> types.GenerateResponse:
    global running_model

    model_type = get_model_type_by_name(name=request.model)
    if running_model is None or running_model.name != model_type.name:
        running_model = model_type()
    assert running_model is not None

    response: types.GenerateResponse = running_model.generate(request=request)
    return response



from . import apis
from . import types


def test_register_model_type():
    class TestModel(types.ModelBase):
        name = "test:latest"

    apis.register_model_type(TestModel)
    assert TestModel in apis.registered_model_types

    assert apis.get_model_type_by_name(name="test") == TestModel



from typing import List
from typing import Optional
from typing import Type
from typing import Union

import numpy as np
import pydantic


class Prompt(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    points: Optional[np.ndarray] = pydantic.Field(default=None)
    point_labels: Optional[np.ndarray] = pydantic.Field(default=None)
    texts: Optional[List[str]] = pydantic.Field(default=None)

    iou_threshold: Optional[float] = pydantic.Field(default=0.5)
    score_threshold: Optional[float] = pydantic.Field(default=0.1)
    max_annotations: Optional[int] = pydantic.Field(default=100)

    @pydantic.field_serializer("points")
    def _serialize_points(self: "Prompt", points: np.ndarray) -> List[List[float]]:
        return points.tolist()

    @pydantic.field_serializer("point_labels")
    def _serialize_point_labels(self: "Prompt", point_labels: np.ndarray) -> List[int]:
        return point_labels.tolist()

    @pydantic.field_validator("points", mode="before")
    @classmethod
    def _validate_points(cls: Type, points: Union[list, np.ndarray]):
        if isinstance(points, list):
            points = np.array(points, dtype=float)
        if points.ndim != 2:
            raise ValueError("points must be 2-dimensional")
        if points.shape[1] != 2:
            raise ValueError("points must have 2 columns")
        return points

    @pydantic.field_validator("point_labels", mode="before")
    @classmethod
    def _validate_point_labels(cls: Type, point_labels: Union[list, np.ndarray]):
        if isinstance(point_labels, list):
            point_labels = np.array(point_labels, dtype=int)
        if point_labels.ndim != 1:
            raise ValueError("point_labels must be 1-dimensional")
        if not set(np.unique(point_labels).tolist()).issubset({0, 1, 2, 3}):
            raise ValueError("point_labels must contain only 0, 1, 2, or 3")
        return point_labels

    @pydantic.model_validator(mode="after")
    def _validate_prompt(self) -> "Prompt":
        if self.points is None and self.point_labels is None:
            if self.texts is None:
                raise ValueError(
                    "texts must be provided when points and point_labels "
                    "are not provided"
                )
        elif (
            self.points is None
            or self.point_labels is None
            or self.points.shape[0] != self.point_labels.shape[0]
        ):
            raise ValueError(
                "points and point_labels must have the same number of rows"
            )
        return self



import numpy as np
import pytest

from . import _json


@pytest.fixture
def image():
    y, x = np.meshgrid(np.arange(10), np.arange(10))
    center = (np.array(x.shape) - 1) / 2
    image_float = np.exp(-((x - center[1]) ** 2 + (y - center[0]) ** 2) / 10)
    image = (image_float * 255).astype(np.uint8)
    return image


def test_image_ndarray_to_b64data(image):
    b64data = _json.image_ndarray_to_b64data(image)
    assert isinstance(b64data, str)
    assert len(b64data) == 204


def test_image_b64data_to_ndarray(image):
    b64data = _json.image_ndarray_to_b64data(image)
    image_recovered = _json.image_b64data_to_ndarray(b64data)
    np.testing.assert_array_equal(image, image_recovered)



import hashlib
from typing import Dict
from typing import Optional

import numpy as np
import onnxruntime
from loguru import logger

from .. import _contextlib
from ._blob import Blob
from ._generate import GenerateRequest
from ._generate import GenerateResponse
from ._image_embedding import ImageEmbedding


class Model:
    name: str

    _blobs: Dict[str, Blob]
    _inference_sessions: Dict[str, onnxruntime.InferenceSession]

    def __init__(self):
        logger.debug("Initializing model {name}", name=self.name)

        self.pull()

        providers = None
        self._inference_sessions = {}
        for key, blob in self._blobs.items():
            try:
                # Try to use all of the available providers e.g., cuda, tensorrt.
                if providers is None:
                    if "CUDAExecutionProvider" in onnxruntime.get_available_providers():
                        providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
                    else:
                        providers = ["CPUExecutionProvider"]
                # Suppress all the error messages from the missing providers.
                with _contextlib.suppress():
                    inference_session = onnxruntime.InferenceSession(
                        blob.path, providers=providers
                    )
            except Exception as e:
                # Even though there is fallback in onnxruntime, it won't always work.
                # e.g., CUDA is installed and CUDA_PATH is set, but CUDA_VISIBLE_DEVICES
                # is empty. We fallback to cpu in such cases.
                logger.error(
                    "Failed to create inference session with providers {providers!r}. "
                    "Falling back to ['CPUExecutionProvider']",
                    providers=providers,
                    e=e,
                )
                providers = ["CPUExecutionProvider"]
                inference_session = onnxruntime.InferenceSession(
                    blob.path, providers=providers
                )
            self._inference_sessions[key] = inference_session

            providers = inference_session.get_providers()
        logger.info(
            "Initialized inference sessions with providers {providers!r}",
            providers=providers,
        )

    @classmethod
    def pull(cls):
        for blob in cls._blobs.values():
            blob.pull()

    @classmethod
    def remove(cls):
        for blob in cls._blobs.values():
            blob.remove()

    @classmethod
    def get_id(cls) -> str:
        return hashlib.md5(
            "+".join(blob.hash for blob in cls._blobs.values()).encode()
        ).hexdigest()[:12]

    @classmethod
    def get_size(cls) -> Optional[int]:
        size = 0
        for blob in cls._blobs.values():
            if blob.size is None:
                return None
            size += blob.size
        return size

    @classmethod
    def get_modified_at(cls) -> Optional[float]:
        modified_at = 0
        for blob in cls._blobs.values():
            if blob.modified_at is None:
                return None
            modified_at = max(modified_at, blob.modified_at)
        return modified_at

    def encode_image(self, image: np.ndarray) -> ImageEmbedding:
        raise NotImplementedError

    def generate(self, request: GenerateRequest) -> GenerateResponse:
        raise NotImplementedError
